"""
TexNew

Automatic LaTeX template management.

:copyright: (c) 2019 by Alex Rutar
:license: MIT, see LICENSE for more details.
"""

__version__ = "1.4"

from texnew.test import parse_errors, test
from texnew.template import build, update
from texnew.document import Document, Divider, TexnewDocument





# print a divider to the specified output
def write_div(out, name):
    out.write("\n" + get_div(name))

# creates a matching regex for file substitution
def repl_match(name):
    if name == "any":
        return r"<\+.*\+>"
    else:
        return r"<\+" + str(name) + r"\+>"
# the main file-building function
def run(target,data,sub_list={},user_macros={}):
    tex_doctype = re.sub(repl_match("doctype"), data['doctype'], read_file("share","defaults","doctype.tex",method="str"))
    tex_packages = read_file("share","defaults","packages.tex",method="str")
    tex_macros = read_file("share","defaults","macros.tex",method="str")
    tex_formatting = read_file("share","formatting",data['formatting'] + '.tex',method="str")
    
    # substitute sub_list
    for k in sub_list.keys():
        tex_formatting = re.sub(repl_match(k), str(sub_list[k]), tex_formatting)

    # generate output file
    with open(target,"a+") as output:
        output.write("% Template created by texnew (author: Alex Rutar); info can be found at 'https://github.com/alexrutar/texnew'.\n")
        output.write("% version ({})\n".format(__version__))

        # create doctype
        write_div(output, "doctype")
        output.write(tex_doctype)

        # add default packates
        write_div(output, "packages")
        output.write(tex_packages)

        # add included macros
        write_div(output, "default macros")
        output.write(tex_macros)
        for name in data['macros']:
            write_div(output, name+" macros")
            output.write(read_file("share","macros",name + ".tex",method="str"))

        # add space for user macros
        write_div(output, "file-specific macros")
        if 'macros' in user_macros.keys():
            for l in user_macros['macros']:
                output.write(l)
        else:
            output.write("% REPLACE\n")

        # add formatting file
        write_div(output, "formatting")
        output.write(tex_formatting)

        # check for contents in user_macros to fill document
        write_div(output, "document start")
        if 'contents' in user_macros.keys():
            for l in user_macros['contents']:
                output.write(l)
        else:
            output.write("\nREPLACE\n")
            output.write("\\end{document}\n")



class TexnewError(Exception):
    pass

class TexnewFileError(TexnewError):
    def __init__(self, filename):
        self.filename = filename
        self.context = "Texnew"
        self.context_info = {}

    def __str__(self):
        return "[{}]: Missing file/directory at \"{}\"".format(self.context,self.filename)

# TODO: catch "IsADirectoryError"
class TexnewInputError(TexnewError):
    def __init__(self, user_input):
        self.user_input = user_input
        self.context = None



# in script: check that target exists
    # print("Error: The file \"{}\" already exists. Please choose another filename.".format(target))
# must specify user file
    # print("Warning: user info file could not be found at 'user.yaml' or at 'user_private.yaml'."), and pass false, if they don't exist
    # check for user file first, raise warning if did not fine
import os
import sys

from .template import build, update, load_template, load_user, available_templates
from .document import TexnewDocument
from .error import TexnewFileError, TexnewInputError
from .file import get_version, get_name

# TODO: have defaults file in .texnew/defaults
def run(fname, template_type):
    # load and catch basic errors with template choice
    if os.path.exists(fname):
        print("Error: The file \"{}\" already exists. Please choose another filename.".format(fname))
        sys.exit(1)

    try:
        user_info = load_user("private")
    except TexnewFileError:
        try:
            user_info = load_user()
        except TexnewFileError as e:
            print("Warning: no user file found, no substitutions will be made!")
            user_info = {}

    try:
        template_data = load_template(template_type)
    except TexnewFileError as e:
        if e.context == "template":
            print("The template \"{}\" does not exist! The possible template names are:\n".format(e.context_info['type'])+ "\t".join(available_templates()))
        else:
            print("uhh unknown error report this")
        sys.exit(1)

    try:
        tdoc = build(template_data, user_info)
        tdoc.write(fname)

    # TODO: improve error handling
    except TexnewFileError as e:
        print(e)
        sys.exit(1)
    except TexnewInputError as e:
        print(e)
        sys.exit(1)

#  def update(tdoc, template_type, transfer=['file-specific preamble', 'document start']):
# TODO: add error handling here
def run_update(fname, template_type):
    # basic checks
    if not os.path.exists(fname):
        print("Error: No file named \"{}\" to update!".format(fname))
        sys.exit(1)
    ver = get_version(fname)
    if ver.startswith("0."):
        print("Error: File too outdated! Must be generated with version at least 1.0; file version is ({})".format(ver))
        sys.exit(1)

    # load the document
    tdoc = TexnewDocument()
    tdoc.load(fname)

    # generate replacement document
    new_tdoc = update(tdoc, template_type)

    # copy the existing file to a new location
    name = get_name(fname,"_old")
    os.rename(fname,name)

    new_tdoc.write(fname)



import os
import yaml
import re
from os.path import expanduser

from . import __path__
from .error import TexnewFileError, TexnewInputError

# get the relative path, from this script
def rpath(*rel_path):
    return os.path.join(expanduser("~"),".texnew",*rel_path)

# methods to open files with special error handling
def read_file(*rel_path, method = "lst", src = "texnew"):
    if src == "texnew":
        path = rpath(*rel_path)
    elif src == "user" and len(rel_path) == 1:
        path = rel_path[0]

    if method == "yaml":
        path += ".yaml"
    try:
        with open(path,'r') as f:
            if method == "str":
                return f.read()
            elif method == "yaml":
                return yaml.load(f)
            elif method == "lst":
                return list(f)
    except FileNotFoundError:
        if src == "texnew":
            raise TexnewFileError(path)
        elif src == "user":
            raise TexnewInputError(path)

# method to get internal list of directories
def get_flist(*rel_path):
    path = rpath(*rel_path)
    try:
        return os.listdir(path)
    except FileNotFoundError:
        e = TexnewFileError(path)
        e.context = "directory"
        raise e

# check for file version
def get_version(filename):
    st = read_file(filename, method = "str", src = "user")
    pat = re.compile(r"% version \((.*)\)")
    res = pat.search(st)
    if res:
        return res.group(1)
    else:
        return "0.1"

# create a new block divider
def get_div(name):
    return ("% " + name + " ").ljust(80, "-") + "\n"

# remove the file endings at rel_path
def truncated_files(*rel_path):
    return ["".join(s.split(".")[:-1]) for s in get_flist(*rel_path)]

# clean the directory at a relative path
def clean_dir(*rel_path):
    for fl in get_flist(*rel_path):
        if not fl.startswith("."):
            os.remove(rpath(*rel_path,fl))

# copy a file to the target; only appends, does not overwrite
def copy_file(src,trg):
    with open(src,'r') as f, open(trg,'a+') as output:
        for l in f:
            output.write(l)

# get an available name, inserting 'ad' if necessary
def get_name(name,ad):
    if "." in name:
        base = "".join(name.split(".")[:-1])
        ftype = name.split(".")[-1]
    else:
        base = name
        ftype = ""
    for t in [""] + ["_"+str(x) for x in range(1000)]:
        attempt = base + ad + t + "." + ftype
        if not os.path.exists(attempt):
            return attempt



import re
import itertools

from .file import read_file
from . import __version__

def _pairwise(iterable):
    "s -> (s0,s1), (s1,s2), (s2, s3), ..."
    a, b = itertools.tee(iterable)
    next(b, None)
    return zip(a, b)

def _strip_block(clist):
    """Removes trailing whitespace from an input list for a block"""
    end = -1
    for i,e in reversed(list(enumerate(clist))):
        if e:
            end = i
            break
    start = 0
    for i,e in enumerate(clist):
        if e:
            start = i
            break
    return clist[start:end+1]

class Divider:
    """Divider class primarily to be used in the document class"""
    def __init__(self, start_symb, fill, length=80):
        if len(fill) != 1:
            raise ValueError("fill must be a string of length 1")
        self.fill = fill
        self.start = start_symb
        self.length = length
    def gen(self, name):
        if len(name) >= self.length:
            raise ValueError("divider name is too long")
        return (self.start + " " + name + " ").ljust(self.length, self.fill)
    def is_div(self, test):
        test = test.rstrip()
        return test.startswith(self.start) and test.endswith(self.fill) and len(test) == self.length
    def name(self, div):
        pat = re.compile("{} (.*) {}*".format(self.start, self.fill))
        res = pat.search(div)
        if res:
            return res.group(1)
        else:
            return None

class Document:
    """A custom method that emulates the python 'dict', but with different construction methods and an inherent order in the keys.
    Also has string methods to appear as a proper block-based document

    sub_list: a dict of possible substitutions to make
    defaults: a dict of fallback values when passed a 'Falsey' value for a block
    div_func: the divider used when printing the Document
    buf: number of newlines to place at end of printing block
    """
    def __init__(self, sub_list={}, defaults={}, div_func=None,buf=0):
        self.div = div_func
        self.subs = sub_list
        self._blocks = {}
        self._order = [] # order matters here!
        self.buf=buf
        self.defaults = defaults

    # return a representation of this object (_blocks and blocks)
    def __repr__(self):
        return "Blocks:\n"+repr(self._blocks) + "\nOrder:\n" + repr(self._order)

    # return a string of this object (e.g. for printing)
    def __str__(self):
        output = ""
        for block in self._order:
            if self.div:
                output += self.div.gen(block) + "\n"
            output += "\n".join(self._blocks[block]) + "\n"*(self.buf+1)
        return output 

    # write to a document
    # TODO: error handling here, perhaps in some more generic write method
    def write(self,fname):
        with open(fname,'a+') as f:
            f.write(str(self))

    # access document indices as blocks
    def __getitem__(self,bname):
        return self._blocks[bname]

    # emulate python dict.get
    def get(self, bname, rep=[]):
        if bname in self._order:
            return self._blocks[bname]
        else:
            return rep

    # check has block
    def __contains__(self,bname):
        return bname in self._blocks

    # add _blocks, will replace if it already exists
    def __setitem__(self, bname, content_list):
        # can input blank content_list in any 'False' format
        if not content_list:
            content_list = self.defaults.get(bname,[])

        # remove trailing whitespace, starting and ending blank lines
        content_list = _strip_block([l.rstrip() for l in content_list])
        
        # substitute matches in content_list with sub_list
        repl_match = lambda x: r"<\+" + str(x) + r"\+>"
        for k in self.subs.keys():
            content_list = [re.sub(repl_match(k), str(self.subs[k]), l) for l in content_list]

        # add _blocks, blocks; overwrites
        self._blocks[bname] = content_list
        if bname not in self._order:
            self._order += [bname]
    
    # returns empty block if not contained
    def __missing__(self,bname):
        return []

    # delete block
    def __delitem__(self, bname):
        del self._blocks[bname]
        self._order.remove(bname)

class TexnewDocument(Document):
    """A special class of type Document with custom loading and checking types, along with a LaTeX style block delimiter"""
    def __init__(self, sub_list={},defaults={}):
        # create default settings when inputting block (if block is none)
        new_defs = {
            'header': ["% Template created by texnew (author: Alex Rutar); info can be found at 'https://github.com/alexrutar/texnew'.",
                    "% version ({})".format(__version__)
                ],
            'file-specific preamble': ["% REPLACE"],
            'document start': ["REPLACE", "\\end{document}"]
        }
        super().__init__(sub_list, div_func=Divider("%","-"), defaults={**new_defs,**defaults}, buf=2)

    # loads a file and appends blocks to current block list
    def load(self,target):
        fl = read_file(target,src="user")
        # read the dividers
        divs = [(i,self.div.name(l)) for i,l in enumerate(fl) if self.div.is_div(l)]

        # break at dividers
        for f,g in _pairwise(divs):
            self[f[1]] =  fl[f[0]+1:g[0]-1]
        self[divs[-1][1]] = fl[divs[-1][0]+1:]
    # TODO: will eventually create an error verification script which consumes a LaTeXDocumet to check that it is an (error-free) LaTeX document
    # TODO: write comparison methods ?



"""
Entry point for the texnew script installed along with this package.



    texnew [-uv] [--user=USER] target template

    banditvis [-cl]
    banditvis -h | --help
    banditvis -V | --version

Positional:
  target                The name of the file to action.
  template              The name of the template to use.

Named:
  --user                Specify the user file to use.

Optional:
  -u                    Update the target template with the given file.
  -v                    Run verbose.

Other:
  -c                    Check existing templates for errors.
  -l                    List existing templates.
"""
import sys
import argparse

from . import __version__
from .test import test
from .scripts import run, run_update
from .template import available_templates
from .file import rpath

def get_usage():
    return '\n\n\n'.join(__doc__.split('\n\n\n')[1:])
# main argument parser, after pre-checking info
def parse():
    parser = argparse.ArgumentParser(prog="texnew",description='An automatic LaTeX template creator and manager.',usage=get_usage())
    parser.add_argument('target', metavar='output', type=str, nargs=1,
                                help='the name of the file you want to create')
    parser.add_argument('template_type', metavar='template', type=str, nargs=1,
                                help='the name of the template to use')

    parser.add_argument('-l', "--list", action="store_true", default=False, dest="lst",help="list existing templates and root folder")
    parser.add_argument('-c', "--check", action="store_true", default=False, dest="lst",help="check for errors in existing templates")
    parser.add_argument('-u', "--update_file", action="store_true", default=False, dest="update_file",help="update_file the specified file with the desired template")
    parser.add_argument("--user", dest="user", default="",help="specify the user file",nargs=1)

    args = parser.parse_args()
    return (args.target[0], args.template_type[0], args.update_file, args.user)

def main():
    # special use cases:
    if ("-h"  in sys.argv[1:]) or ("--help" in sys.argv[1:]) or (len(sys.argv) == 1):
        print(get_usage())
    elif "-V" in sys.argv[1:] or "--version" in sys.argv[1:]:
        print("texnew ({})".format(__version__))
    elif "-l" in sys.argv[1:]:
        print("\nRoot Folder: {}/".format(rpath()))
        print("Existing templates:\n"+ "\t".join(available_templates()))
    elif "-c" in sys.argv[1:]:
        test()
    else:
        # main program branching
        target, template_type, update_file, user= parse()
        if update_file:
            run_update(target, template_type)
        else:
            if not target.endswith(".tex"):
                target = target + ".tex"
            run(target, template_type)

# entry point for script
if __name__ == "__main__":
    main()



import re

from .file import read_file, truncated_files
from .error import TexnewFileError
from .document import TexnewDocument

def load_template(template_type):
    """Load template information for template_data"""
    try:
        return read_file("templates",template_type,method="yaml")
    except TexnewFileError as e:
        e.context = "template"
        e.context_info['type'] = template_type
        raise e

# TODO: write this function
def available_templates():
    return truncated_files("templates")

def load_user(info_name = "default"):
    """Load user information for sub_list"""
    try:
        return read_file("user",info_name,method="yaml")
    except TexnewFileError as e:
        e.context = "user"
        e.context_info['name'] = info_name
        raise e

def build(template_data, sub_list={}):
    """Build a TexnewDocument from existing template_data.
    Note: makes a lot of assumptions about the structure of template_data"""
    sub_list['doctype'] = template_data['doctype']
    tdoc = TexnewDocument(sub_list)
    rel = ['share', template_data['template']]

    # set default header
    tdoc['header'] = None

    # default components
    tdoc['doctype'] =  read_file(*rel,"defaults","doctype.tex")
    tdoc['packages'] =  read_file(*rel,"defaults","packages.tex")
    tdoc['default macros'] =  read_file(*rel,"defaults","macros.tex")

    # special macros
    for name in template_data['macros']:
        tdoc['macros ({})'.format(name)] = read_file(*rel,"macros",name + ".tex")
    
    # add existing user macros
    tdoc['file-specific preamble'] =  None

    # formatting block
    tdoc['formatting'] = read_file(*rel,"formatting",template_data['formatting']+ ".tex")

    # add existing user contents
    tdoc['document start'] = None

    return tdoc

def update(tdoc, template_type, transfer=['file-specific preamble', 'document start']):
    # generate replacement document
    user_info = load_user()
    template_data = load_template(template_type)
    new_tdoc = build(template_data, sub_list=user_info)

    # write information to new document
    for bname in transfer:
        new_tdoc[bname] = tdoc[bname]

    return new_tdoc
